from sys import platform as sys_pf
import os
import antimony
import roadrunner
import rrplugins
import scipy
import sympy
import numpy.linalg
import time
import math
import shutil
if sys_pf == 'darwin':
    import matplotlib
    matplotlib.use("TkAgg")
    import matplotlib.pyplot as plt
else:
    import matplotlib.pyplot as plt
roadrunner.Logger.setLevel(roadrunner.Logger.LOG_ERROR)
rrplugins.setLogLevel('error')


class BistabilityFinder:

    @classmethod
    def run_optimization(cls, bounds, num_constraint_method_iters, sys_min_val, temp_c, penalty_objective_func,
                         feasible_point_check, objective_function_to_optimize, final_constraint_check, seed,
                         equality_bounds_indices, print_flag, numpy_dtype, concentration_bounds):
        important_info = ''

        # running penalty method to find feasible points
        print("")
        print("Running feasible point method for " + str(num_constraint_method_iters) + " iterations ...")
        start = time.process_time()
        x_candidates = cls.feasible_point_method(bounds, num_constraint_method_iters, sys_min_val, temp_c,
                                                 penalty_objective_func, feasible_point_check, seed,
                                                 equality_bounds_indices, print_flag,
                                                 numpy_dtype, concentration_bounds)
        end = time.process_time()
        print("Elapsed time for feasible point method: " + str(end - start))
        print("")

        num_x_cand = len(x_candidates)
        important_info += "The number of feasible points that satisfy the constraints: " + str(num_x_cand) + "\n"

        # if num_x_cand = 0 throw error and don't cont.
        if num_x_cand != 0:
            print("Running the multistart optimization ...")
            start = time.process_time()

            params_that_give_global_min, total_that_give_zero, total_that_pass_chk, obj_fun_val_global_min = \
                cls.multistart_optimization(num_x_cand, x_candidates, sys_min_val, bounds, temp_c,
                                            objective_function_to_optimize, equality_bounds_indices, seed,
                                            final_constraint_check, print_flag, numpy_dtype, concentration_bounds)
            
            end = time.process_time()
            print("Elapsed time for multistart method: " + str(end - start))
            print("")

            important_info += "Total feasible points that give F(x) = 0: " + str(total_that_give_zero) + "\n"

            important_info += "Total number of points that passed final_check: " + str(total_that_pass_chk) + "\n"

            return params_that_give_global_min, obj_fun_val_global_min, important_info
        else:
            raise Exception("Optimization needs to be run with more iterations or different bounds.")

    @classmethod
    def feasible_point_method(cls, penalty_bounds, num_constraint_method_iters, sys_min_val, temp_c,
                              penalty_objective_func, feasible_point_check, seed, equality_bounds_indices, print_flag,
                              numpy_dtype, concentration_bounds):
        # Generate starting points uniformly with length ranges
        numpy.random.seed(seed)
        samples = numpy.random.rand(num_constraint_method_iters, len(penalty_bounds) -
                                    len(equality_bounds_indices)).astype(numpy_dtype)

        x_candidates = []
        x_off = numpy.zeros(len(penalty_bounds), dtype=numpy_dtype)

        non_equality_bounds_indices = [i for i in range(len(penalty_bounds)) if i not in equality_bounds_indices]
        
        true_bounds = [(numpy_dtype(penalty_bounds[j][0]), numpy_dtype(penalty_bounds[j][1]))
                       for j in non_equality_bounds_indices]

        ranges = numpy.asarray(true_bounds, dtype=numpy_dtype)
        samples = samples*(ranges[:, 1] - ranges[:, 0]) + ranges[:, 0]

        for n in range(num_constraint_method_iters):
            with numpy.errstate(divide='ignore', invalid='ignore'):
                result = scipy.optimize.minimize(penalty_objective_func, samples[n], args=
                (temp_c, penalty_bounds, equality_bounds_indices, x_off, non_equality_bounds_indices,
                 concentration_bounds), method='SLSQP', tol=1e-16, bounds=true_bounds)

                if abs(result.fun) > numpy_dtype(1e-100): 
                    result0 = scipy.optimize.minimize(penalty_objective_func, result.x, args=
                    (temp_c, penalty_bounds, equality_bounds_indices, x_off, non_equality_bounds_indices,
                     concentration_bounds), method='Nelder-Mead', tol=1e-16)

                    output = feasible_point_check(result0.x, result0.fun, sys_min_val, equality_bounds_indices,
                                                  non_equality_bounds_indices, penalty_bounds, concentration_bounds)
                    if print_flag:
                        print("Objective function value: " + str(result0.fun))
                        print("Decision vector used: ")
                        print(result0.x)
                        print("")

                    if output:
                        x_candidates.append(result0.x)
                else:
                    output = feasible_point_check(result.x, result.fun, sys_min_val, equality_bounds_indices,
                                                  non_equality_bounds_indices, penalty_bounds, concentration_bounds)
                    if print_flag:
                        print("Objective function value: " + str(result.fun))
                        print("Decision vector used: ")
                        print(result.x)
                        print("")
                    if output:
                        x_candidates.append(result.x)

        return x_candidates

    @classmethod 
    def multistart_optimization(cls, num_x_cand, x_candidates, sys_min_val, penalty_bounds, temp_c,
                                objective_function_to_optimize, equality_bounds_indices, seed_given,
                                final_constraint_check, print_flag, numpy_dtype, concentration_bounds):
        x_off = numpy.zeros(len(penalty_bounds), dtype=numpy_dtype)
        non_equality_bounds_indices = [i for i in range(len(penalty_bounds)) if i not in equality_bounds_indices]

        smallest_value = numpy_dtype(numpy.PINF)
        x_that_give_global_min = []
        obj_fun_val_global_min = []
        start = 0
        for i in range(num_x_cand):
            with numpy.errstate(divide='ignore', invalid='ignore'):
                result = scipy.optimize.basinhopping(objective_function_to_optimize, x_candidates[i],
                                                     minimizer_kwargs={'method': 'Nelder-Mead',
                                                                       'args': (temp_c, penalty_bounds, sys_min_val,
                                                                                equality_bounds_indices, x_off,
                                                                                non_equality_bounds_indices,
                                                                                concentration_bounds), 'tol': 1e-16},
                                                     niter=2, seed=seed_given)

                start = start+1

                if print_flag:
                    print("Global function value: " + str(result.fun))
                    print("Decision vector produced: ")
                    print(result.x)
                
                if result.fun > numpy_dtype(1e-200):
                    result = scipy.optimize.minimize(objective_function_to_optimize, result.x,
                                                     args=(temp_c, penalty_bounds, sys_min_val, equality_bounds_indices,
                                                           x_off, non_equality_bounds_indices, concentration_bounds),
                                                     method='Nelder-Mead', tol=1e-16)
        
                    if print_flag:
                        print("Local function value: " + str(result.fun))
                        print("Decision vector produced: ")
                        print(result.x)

                if print_flag:
                    print("")
                        
                if result.fun < smallest_value:
                    smallest_value = result.fun

            if abs(result.fun) <= sys_min_val: 
                x_that_give_global_min.append(result.x)
                obj_fun_val_global_min.append(result.fun)

        x_that_give_global_min2, total_that_give_zero, total_that_pass_chk, obj_fun_val_global_min2 = \
            cls.__create_final_points(x_that_give_global_min, obj_fun_val_global_min, final_constraint_check,
                                      penalty_bounds, sys_min_val, equality_bounds_indices, concentration_bounds)

        print("\nSmallest value achieved by objective function: " + str(smallest_value) + "\n")

        return x_that_give_global_min2, total_that_give_zero, total_that_pass_chk, obj_fun_val_global_min2

    @staticmethod
    def __create_final_points(x_that_give_global_min, obj_fun_val_global_min, final_constraint_check, penalty_bounds,
                              sys_min_val, non_equality_bounds_indices, concentration_bounds):
        if len(x_that_give_global_min) != 0:
            temp2 = numpy.vstack(x_that_give_global_min)
            u, indices = numpy.unique(temp2, axis=0, return_index=True)
            x_that_give_global_min2 = []
            obj_fun_val_global_min2 = []
            for i in indices:

                output = final_constraint_check(x_that_give_global_min[i], penalty_bounds, sys_min_val,
                                                non_equality_bounds_indices, concentration_bounds)

                if output[0]:
                    x_that_give_global_min2.append(output[1])
                    obj_fun_val_global_min2.append(obj_fun_val_global_min[i])
            total_that_give_zero = len(indices)
            total_that_pass_chk = len(x_that_give_global_min2)
        else:
            x_that_give_global_min2 = []
            obj_fun_val_global_min2 = []
            total_that_give_zero = 0
            total_that_pass_chk = 0

        return x_that_give_global_min2, total_that_give_zero, total_that_pass_chk, obj_fun_val_global_min2
        
    @classmethod
    def run_continuity_analysis(cls, species_num, params_for_global_min, initialize_ant_string, finalize_ant_string,
                                species_y, dir_path, print_lbls_flag, auto_parameters):

        print("Running continuity analysis ...")
        init_ant, pcp_x = initialize_ant_string(species_num, auto_parameters['PrincipalContinuationParameter'])
        auto_parameters['PrincipalContinuationParameter'] = pcp_x
        start = time.process_time()
        multistable_param_ind = []
        auto = rrplugins.Plugin("tel_auto2000")
        cont_direction = ["Positive", "Negative"]

        for param_ind in range(len(params_for_global_min)):
            final_ant_str = finalize_ant_string(params_for_global_min[param_ind], init_ant)

            for dir_ind in range(2):
                if os.path.isdir("./auto_fort_files"):
                    shutil.rmtree("./auto_fort_files")

                pts, lbls, bi_data_np, antimony_r, flag = cls.run_numerical_continuation\
                    (final_ant_str, cont_direction[dir_ind], auto, auto_parameters)

                if print_lbls_flag:
                    print("Labels from numerical continuation: ")
                    print(lbls)
                
                if flag and lbls != ['EP', 'EP']:
                    chnk_stable, chnk_unstable, special_points, sp_y_ind = cls.find_stable_unstable_regions(antimony_r,
                                                                                                            species_y)

                    multistable = cls.detect_multi_stability(chnk_stable, chnk_unstable, bi_data_np)

                    if multistable:
                        cls.plot_pcp_vs_species(chnk_stable, chnk_unstable, special_points, bi_data_np, sp_y_ind, pcp_x,
                                                species_y, param_ind, dir_path, cls)
                        multistable_param_ind.append(param_ind)
                        break

        if os.path.isdir("./auto_fort_files"):
            shutil.rmtree("./auto_fort_files")

        end = time.process_time()
        print("Elapsed time for continuity analysis: " + str(end - start))
        print("")

        important_info = "Number of multistability plots found: " + str(len(multistable_param_ind)) + "\n"
        
        important_info += "Elements in params_for_global_min that produce multistability: \n" + \
                          str(multistable_param_ind) + "\n"

        return multistable_param_ind, important_info

    @classmethod
    def run_greedy_continuity_analysis(cls, species_num, params_for_global_min, initialize_ant_string,
                                       finalize_ant_string,
                                       species_y, dir_path, print_lbls_flag, auto_parameters):

        print("Running continuity analysis ...")
        init_ant, pcp_x = initialize_ant_string(species_num, auto_parameters['PrincipalContinuationParameter'])
        auto_parameters['PrincipalContinuationParameter'] = pcp_x
        auto_parameters['IADS'] = 0
        auto_parameters['A1'] = 1e10
        auto_parameters['ITNW'] = 100
        auto_parameters['NTST'] = 100
        auto_parameters['NCOL'] = 100

        start = time.process_time()
        multistable_param_ind = []
        auto = rrplugins.Plugin("tel_auto2000")
        cont_direction = ["Positive", "Negative"]
        for param_ind in range(len(params_for_global_min)):

            final_ant_str = finalize_ant_string(params_for_global_min[param_ind], init_ant)
            pcp_ranges_mag = cls.get_pcp_ranges(final_ant_str, pcp_x)

            auto_parameters['RL0'] = pcp_ranges_mag[0]
            auto_parameters['RL1'] = pcp_ranges_mag[1]

            ds_vals = []
            mag = pcp_ranges_mag[2]
            mag -= 1
            for i in range(5):
                ds_vals.append(float(10 ** mag))
                mag -= 1

            multistable = False
            lbls = []
            for ds_val in ds_vals:
                auto_parameters['DS'] = ds_val
                for dir_ind in range(2):
                    if os.path.isdir("./auto_fort_files"):
                        shutil.rmtree("./auto_fort_files")

                    pts, lbls, bi_data_np, antimony_r, flag = cls.run_numerical_continuation \
                        (final_ant_str, cont_direction[dir_ind], auto, auto_parameters)

                    if print_lbls_flag:
                        print("Labels from numerical continuation: ")
                        print(lbls)

                    if flag and lbls != ['EP', 'EP']:
                        chnk_stable, chnk_unstable, special_points, sp_y_ind = cls.find_stable_unstable_regions(
                            antimony_r,
                            species_y)

                        multistable = cls.detect_multi_stability(chnk_stable, chnk_unstable, bi_data_np)

                        if multistable:

                            # running another numerical continuation with a smaller step size to try and get
                            # better looking plots
                            scnd_check = True
                            if os.path.isdir("./auto_fort_files"):
                                shutil.rmtree("./auto_fort_files")

                                ind_ds_val = ds_vals.index(ds_val)
                                ds_val = ds_vals[ind_ds_val + 1]

                                auto_parameters['DS'] = ds_val
                                for dir_ind2 in range(2):
                                    if os.path.isdir("./auto_fort_files"):
                                        shutil.rmtree("./auto_fort_files")

                                    pts2, lbls2, bi_data_np2, antimony_r, flag2 = cls.run_numerical_continuation \
                                        (final_ant_str, cont_direction[dir_ind2], auto, auto_parameters)

                                    if flag2 and lbls2 != ['EP', 'EP']:
                                        chnk_stable2, chnk_unstable2, special_points2, sp_y_ind2 = \
                                            cls.find_stable_unstable_regions(antimony_r, species_y)

                                        multistable2 = cls.detect_multi_stability(chnk_stable2, chnk_unstable2,
                                                                                  bi_data_np2)

                                        if multistable2:
                                            cls.plot_pcp_vs_species(chnk_stable2, chnk_unstable2, special_points2,
                                                                    bi_data_np2, sp_y_ind2,
                                                                    pcp_x, species_y, param_ind, dir_path, cls)
                                            scnd_check = False
                                            break

                                if scnd_check:
                                    cls.plot_pcp_vs_species(chnk_stable, chnk_unstable, special_points,
                                                            bi_data_np, sp_y_ind, pcp_x, species_y, param_ind,
                                                            dir_path, cls)

                            if param_ind not in multistable_param_ind:
                                multistable_param_ind.append(param_ind)
                            break
                if multistable and 'MX' not in lbls:
                    break
            if print_lbls_flag:
                print("")

        if os.path.isdir("./auto_fort_files"):
            shutil.rmtree("./auto_fort_files")

        end = time.process_time()
        print("Elapsed time for continuity analysis: " + str(end - start))
        print("")

        important_info = "Number of multistability plots found: " + str(len(multistable_param_ind)) + "\n"

        important_info += "Elements in params_for_global_min that produce multistability: \n" + \
                          str(multistable_param_ind) + "\n"

        return multistable_param_ind, important_info

    @staticmethod
    def get_pcp_ranges(final_ant_str, pcp_x):

        splits = final_ant_str.split(';')
        initial_vars = [i for i in splits if '=' in i]
        symbol_and_vals = []
        for j in initial_vars:
            symbol_and_vals.append([i.strip() for i in j.split('=')])

        pcp_initial = [symbol_and_vals[i] for i in range(len(symbol_and_vals))
                       if symbol_and_vals[i][0] == pcp_x][0][1]

        def order_of_magnitude(number):
            return math.floor(math.log(number, 10))

        mag = order_of_magnitude(float(pcp_initial))

        for i in pcp_initial:
            if i != '0' and i != '.':
                val = i
                break

        rl1 = float(val) * 10 ** (mag + 1)
        rl0 = float(val) * 10 ** (mag - 1)

        return [rl0, rl1, mag + 1]

    @classmethod
    def run_numerical_continuation(cls, ant_str, direction, auto, auto_parameters):
        flag = True
        antimony_r = cls.__loada(ant_str)

        # making the directory auto_fort_files if is does not exist
        if not os.path.isdir("./auto_fort_files"):
            os.mkdir("./auto_fort_files")
        auto.setProperty("SBML", antimony_r.getCurrentSBML())
        auto.setProperty("ScanDirection", direction)
        auto.setProperty("PreSimulation", "True")
        auto.setProperty("PreSimulationDuration", 1.0)
        auto.setProperty('KeepTempFiles', True)
        auto.setProperty("TempFolder", "./auto_fort_files")

        # assigning values provided by the user
        for i in auto_parameters.keys():
            auto.setProperty(i, auto_parameters[i])
        try:

            auto.execute()
            # indices where special points are
            pts = auto.BifurcationPoints
            # labeling of special points
            lbls = auto.BifurcationLabels
            # all data for parameters and species found by continuation
            bi_data = auto.BifurcationData

            # convertes bi_data to numpy array, where first
            # column is the principal continuation parameter and
            # the rest of the columns are the species
            bi_data_np = bi_data.toNumpy

        except Exception as e:
            flag = False
            pts = []
            lbls = []
            bi_data_np = []

        return pts, lbls, bi_data_np, antimony_r, flag

    @staticmethod
    def find_stable_unstable_regions(antimony_r, species_y):
        with open("./auto_fort_files/fort.7", 'r') as fobj:
            all_lines = [[num for num in line.split()] for line in fobj]
        
        # getting index where important information begins
        beg_ind = None
        for i in range(len(all_lines)):
            if all_lines[i][0] == '1':
                beg_ind = i
                break

        solution_types = {'1': 'BP', '2': 'LP', '3': 'HB', '5': 'LP', '6': 'BP', '7': 'PD', '8': 'TR',
                          '9': 'EP', '-9': 'MX'}

        # lists to hold stable and unstable regions, and special points
        unstable = []
        stable = []
        special_points = []
        ignored_labels = ['0', '4', '-4']
        for i in range(beg_ind, len(all_lines)):
            if all_lines[i][0] == '0':
                break
            else:
                temp = all_lines[i][1:3]
                if temp[1] in ignored_labels:
                    if temp[0][0] == "-":
                        stable.append(abs(int(temp[0]))-1)
                    else:
                        unstable.append(int(temp[0])-1)
                else:
                    special_points.append((abs(int(temp[0]))-1, solution_types[temp[1]]))
            
        # getting stable chunks of indices
        spl = [0] + [i for i in range(1, len(stable)) if stable[i] - stable[i-1] > 1]+ [None]
        chnk_stable = [stable[b:e] for (b, e) in [(spl[i-1], spl[i]) for i in range(1, len(spl))]]

        # getting unstable chunks of indices
        spl = [0] + [i for i in range(1, len(unstable)) if unstable[i] - unstable[i-1] > 1] + [None]
        chnk_unstable = [unstable[b:e] for (b, e) in [(spl[i-1], spl[i]) for i in range(1, len(spl))]]

        # species ids according to AUTO also order in biData
        spec_id = antimony_r.model.getFloatingSpeciesIds()

        # index of species_y in biData
        sp_y_ind = spec_id.index(species_y) + 1

        return chnk_stable, chnk_unstable, special_points, sp_y_ind

    @classmethod
    def detect_multi_stability(cls, chnk_stable, chnk_unstable, bi_data_np):

        if cls.is_list_empty(chnk_stable) or cls.is_list_empty(chnk_unstable):
            return False

        chnk_stable_pcp_ranges = []
        for i in range(len(chnk_stable)):
            end_ind = len(chnk_stable[i])-1 
            chnk_stable_pcp_ranges.append([bi_data_np[chnk_stable[i][0], 0], bi_data_np[chnk_stable[i][end_ind], 0]])

        [i.sort() for i in chnk_stable_pcp_ranges] 

        chnk_unstable_pcp_ranges = []
        for i in range(len(chnk_unstable)):
            end_ind = len(chnk_unstable[i])-1
            chnk_unstable_pcp_ranges.append([bi_data_np[chnk_unstable[i][0], 0],
                                             bi_data_np[chnk_unstable[i][end_ind], 0]])

        [i.sort() for i in chnk_unstable_pcp_ranges] 

        chnk_unstable_pcp_intervals = []
        chnk_stable_pcp_intervals = []
        for i in range(len(chnk_unstable)):
            chnk_unstable_pcp_intervals.append(sympy.Interval(chnk_unstable_pcp_ranges[i][0],
                                                              chnk_unstable_pcp_ranges[i][1]))

        for i in range(len(chnk_stable)):
            chnk_stable_pcp_intervals.append(sympy.Interval(chnk_stable_pcp_ranges[i][0],
                                                            chnk_stable_pcp_ranges[i][1]))

        # building intersections of unstable branch with stable branches
        unstable_intersections = []
        for i in chnk_unstable_pcp_intervals:
            unstable_intersections.append([1 for j in chnk_stable_pcp_intervals if i.intersect(j) != sympy.EmptySet()])

        # creating an object to store x axis limits based on intervals
        xlabel_interval_list = []
        for i in chnk_unstable_pcp_intervals:
            xlabel_interval_list.append([i.intersect(j) for j in chnk_stable_pcp_intervals])

        cls.__xlabel_interval = []
        for i in xlabel_interval_list:
            total_interval = sympy.S.EmptySet
            for j in i:
                total_interval = sympy.Union(total_interval, j)
            if total_interval != sympy.EmptySet():

                if not cls.is_list_empty(cls.__xlabel_interval):
                    intersections = [total_interval.intersect(i) for i in cls.__xlabel_interval]
                else:
                    intersections = [sympy.EmptySet()]

                if sympy.EmptySet() in intersections:
                    cls.__xlabel_interval.append(total_interval)

        return any([sum(i) >= 2 for i in unstable_intersections])

    @classmethod
    def is_list_empty(cls, in_list):
        if isinstance(in_list, list):  # Is a list
            return all(map(cls.is_list_empty, in_list))
        return False  # Not a list

    @staticmethod
    def plot_pcp_vs_species(chnk_stable, chnk_unstable, special_points, bi_data_np, sp_y_ind, pcp_x, species_y,
                            param_ind, dir_path, cls):

        # plotting stable points
        for i in range(len(chnk_stable)):
            plt.plot(bi_data_np[chnk_stable[i], 0], bi_data_np[chnk_stable[i], sp_y_ind], linewidth=1,
                     linestyle='-', color='b')

        # plotting unstable points
        for i in range(len(chnk_unstable)):
            plt.plot(bi_data_np[chnk_unstable[i], 0], bi_data_np[chnk_unstable[i], sp_y_ind], linewidth=1,
                     linestyle='--', color='b')

        # plotting the special points with red marker *
        for i in special_points:
            plt.plot(bi_data_np[i[0], 0], bi_data_np[i[0], sp_y_ind], marker='*', color='r')
            plt.annotate(i[1], (bi_data_np[i[0], 0], bi_data_np[i[0], sp_y_ind]))
        
        #count = 0
        margin = 1e-6
        plt.xlabel(pcp_x)
        plt.ylabel(species_y)

        pcp_values = []
        for i in special_points:
            if i[1] != 'EP':
                pcp_values.append(bi_data_np[i[0], 0])

        pcp_min = min(pcp_values)
        pcp_max = max(pcp_values)

        if abs(pcp_min - pcp_max) > 1e-16:
            diff_x = pcp_max - pcp_min
            start = pcp_min - diff_x
            end = pcp_max + diff_x
            xlim_list = [start, end]
            plt.xlim(xlim_list)
        else:
            start = pcp_min - pcp_min*margin
            end = pcp_max + pcp_max*margin
            xlim_list = [start, end]
            plt.xlim(xlim_list)

        lims = plt.gca().get_xlim()
        x = bi_data_np[:, 0]
        y = bi_data_np[:, sp_y_ind]
        i = numpy.where((x > lims[0]) & (x < lims[1]))[0]

        h = y[i].max() - y[i].min()
        first = y[i].min() - 0.1*h
        second = y[i].max() + 0.1*h
        plt.gca().set_ylim(first, second)

        plt.ticklabel_format(axis='both', style='sci', scilimits=(-2, 2))
        plt.savefig(dir_path + '/' + pcp_x + '_vs_' + species_y + '_' + str(param_ind) + '.png') #'.pdf')

        # going through all the multistationary regions
        # for j in cls.__xlabel_interval:
        #     # setting xlim based on multistationary region
        #     if sympy.utilities.iterables.iterable(j):
        #         start = float(list(j)[0] - list(j)[0]*margin)
        #         end = float(list(j)[0] + list(j)[0]*margin)
        #         xlim_list = [start, end]
        #     elif type(j) == sympy.Union:
        #         start = float(j.inf)
        #         end = float(j.sup)
        #         diff_x = start - end
        #         xlim_list = [start - margin * diff_x, end + margin * diff_x]
        #     else:
        #         start = float(j.start)
        #         end = float(j.end)
        #         diff_x = start - end
        #         xlim_list = [start - margin*diff_x, end + margin*diff_x]
        #
        #     plt.xlim(xlim_list)
        #
        #     # reformatting the y axis based on xlim chosen
        #     lims = plt.gca().get_xlim()
        #     x = bi_data_np[:, 0]
        #     y = bi_data_np[:, sp_y_ind]
        #     i = numpy.where((x > lims[0]) & (x < lims[1]))[0]
        #
        #     if len(i) != 0:
        #         if len(i) == 1:
        #             first = y[i] - y[i]*0.03
        #             second = y[i] + y[i]*0.03
        #         else:
        #             h = y[i].max()-y[i].min()
        #             first = y[i].min() - 0.1*h
        #             second = y[i].max() + 0.1*h
        #         plt.gca().set_ylim(first, second)
        #
        #         # setting labels
        #         plt.xlabel(pcp_x)
        #         plt.ylabel(species_y)
        #
        #         plt.ticklabel_format(axis='both', style='sci', scilimits=(-2, 2))
        #         plt.savefig(dir_path + '/' + pcp_x + '_vs_' + species_y + '_' + str(param_ind) + '_' + str(count) + '.pdf')
        #         count += 1

        plt.close()

    # functions taken from Tellurium!! Give them
    # credit, they deserve it!
    #################################################
    @staticmethod
    def __check_antimony_return_code(code):

        if code < 0:
            raise Exception('Antimony: {}'.format(antimony.getLastError()))

    @classmethod
    def __antimony_to_sbml(cls, ant):
        try:
            isfile = os.path.isfile(ant)
        except ValueError:
            isfile = False
        if isfile:
            code = antimony.loadAntimonyFile(ant)
        else:
            code = antimony.loadAntimonyString(ant)
        cls.__check_antimony_return_code(code)
        mid = antimony.getMainModuleName()
        return antimony.getSBMLString(mid)

    @classmethod
    def __loada(cls, ant):

        return cls.__load_antimony_model(ant)

    @classmethod
    def __load_antimony_model(cls, ant):

        sbml = cls.__antimony_to_sbml(ant)
        return roadrunner.RoadRunner(sbml)



import networkx
import numpy
import sympy
import scipy.optimize
import sys
import matplotlib.cbook
import warnings
from sys import platform as sys_pf
if sys_pf == 'darwin':
    import matplotlib
    matplotlib.use("TkAgg")
    import matplotlib.pyplot as plt
else:
    import matplotlib.pyplot as plt

warnings.filterwarnings("ignore", category=matplotlib.cbook.mplDeprecation)


class Cgraph:
    """
    Class for constructing core CRNT values and C-graph of the network.
    """
    def __init__(self, model):
        """
        Initialization of Cgraph class.

        See also
        ---------
        crnt4sbml.CRNT.get_c_graph()
        """
        self.__g = networkx.DiGraph()

        # list of nodes and edges in the order they were added
        self.__g_nodes = []
        self.__g_edges = []

        self.__dic_species_id_bcs = {}
        for i in model.getListOfSpecies():
            self.__dic_species_id_bcs.update({i.getId(): i.getBoundaryCondition()})

        self.__species = [i.getId() for i in model.getListOfSpecies() if not i.getBoundaryCondition()]

        # temporary
        ######################################################################################
        self.__species_name = [i.getName() for i in model.getListOfSpecies()]
        self.__dic_species_id_name = {}
        for i in range(len(self.__species)):
            self.__dic_species_id_name.update({self.__species[i]: self.__species_name[i]})
        ######################################################################################

        self.__parse_reactions(model)

        self.__linkage_classes = [self.__g.subgraph(c) for c in networkx.weakly_connected_components(self.__g)]

        self.__complexes = [self.__g.nodes[n]['label'] for n in self.__g_nodes]
        self.__reactions = [self.__g.edges[e]['label'] for e in self.__g_edges]

        # self.__check_weak_reversibility_of_linkage_classes() # todo: candidate for deletion

        # core calculations
        self.__create_y_matrix()
        self.__create_s_matrix()
        self.__create_a_matrix()
        self.__create_mass_action_monomials()

        if all([i <= 1 for i in self.get_number_of_terminal_strong_lc_per_lc()]) and \
                self.get_dim_equilibrium_manifold() > 0:
            self.__create_b_matrix()
            self.__create_lambda_matrix()

        # self.__calculate_deficiency() # todo: candidate for deletion
        # self.__calculate_number_of_conservation_relations() # todo: candidate for deletion
        # self.__classify_dimensionality() # todo: candidate for deletion

    def __parse_reactants_and_products(self, reaction):

        # extracting reactants
        reactants = [i.getSpecies() for i in reaction.getListOfReactants()]

        bound_cond = any([self.__dic_species_id_bcs.get(n, n) for n in reactants])
        if bound_cond:
            color = 'red'
        else:
            color = 'green'

        stoichiometries = [int(i.getStoichiometry()) for i in reaction.getListOfReactants()]
        temp = zip([repr(i) if i != 1 else '' for i in stoichiometries], reactants)
        reactant_complex_name = '+'.join(['*'.join(filter(lambda x: x != '', i)) for i in temp])
        if self.__g.number_of_nodes() == 0 or reactant_complex_name not in [self.__g.nodes[v]['label'] for
                                                                            v in self.__g.nodes]:
            self.__g.add_node(reactant_complex_name,
                              label=reactant_complex_name,
                              species=reactants,
                              stoichiometries=stoichiometries,
                              species_bc=bound_cond,
                              color=color)
            self.__g_nodes.append(reactant_complex_name)

        # extracting products
        products = [i.getSpecies() for i in reaction.getListOfProducts()]

        bound_cond = any([self.__dic_species_id_bcs.get(n, n) for n in products])
        if bound_cond:
            color = 'red'
        else:
            color = 'green'

        stoichiometries = [int(i.getStoichiometry()) for i in reaction.getListOfProducts()]
        temp = zip([repr(i) if i != 1 else '' for i in stoichiometries], products)
        product_complex_name = '+'.join(['*'.join(filter(lambda x: x != '', i)) for i in temp])
        if product_complex_name not in [self.__g.nodes[v]['label'] for v in self.__g.nodes]: 
            self.__g.add_node(product_complex_name,
                              label=product_complex_name,
                              species=products,
                              stoichiometries=stoichiometries,
                              species_bc=bound_cond,
                              color=color)
            self.__g_nodes.append(product_complex_name)

        return [reactant_complex_name, product_complex_name]

    def __extract_direct_reaction(self, reaction):
        r, p = self.__parse_reactants_and_products(reaction)
        self.__g.add_edge(r, p,
                          label=reaction.getId(),
                          k=None,
                          sbml_label=reaction.getId())
        self.__g_edges.append((r, p))

    def __extract_reverse_reaction(self, reaction):
        r, p = self.__parse_reactants_and_products(reaction)
        self.__g.add_edge(p, r,
                          label=reaction.getId() + 'r',
                          k=None,
                          sbml_label=reaction.getId())
        self.__g_edges.append((p, r))

    def __parse_reactions(self, model):
        for i in model.getListOfReactions():
            self.__extract_direct_reaction(i)
            if i.getReversible():
                self.__extract_reverse_reaction(i)

    # core CRNT calcs
    def __create_y_matrix(self):

        self.__Y = sympy.zeros(len(self.__species), len(self.__complexes))
        count = 0
        for i in self.__g_nodes:
            v = self.__g.nodes[i]
            stoich = dict(zip(v['species'], v['stoichiometries']))
            self.__Y[:, count] = [stoich[i] if i in stoich.keys() else 0 for i in self.__species]
            count += 1
            
    def __create_s_matrix(self):

        self.__S = sympy.zeros(len(self.__species), len(self.__reactions))
        count = 0 
        for i in self.__g_edges:
            target = self.__complexes.index(i[1])
            source = self.__complexes.index(i[0])
            self.__S[:, count] = self.__Y[:, target] - self.__Y[:, source]
            count += 1

    def __create_lambda_matrix(self):
        self.__Lambda = sympy.zeros(len(self.__complexes), len(self.__linkage_classes))
        for li in range(len(self.__linkage_classes)):
            l_complexes = [i for i in self.__linkage_classes[li].nodes]
            self.__Lambda[:, li] = [1 if i in l_complexes else 0 for i in self.__complexes]

    def __create_a_matrix(self):
        self.__A = sympy.zeros(len(self.__complexes), len(self.__complexes))
        inc_list = [list(self.__g.out_edges(n, data='label')) for n in self.__g_nodes]

        for i in range(len(self.__complexes)):
            for ii in inc_list[i]:
                self.__A[i, i] = self.__A[i, i] - sympy.Symbol(ii[2], positive=True)

        for i in range(len(self.__complexes)):
            for j in range(len(self.__complexes)):
                if self.__g.has_edge(self.__complexes[j], self.__complexes[i]):
                    self.__A[i, j] = self.__A[i, j] + sympy.Symbol(self.__g.get_edge_data(self.__complexes[j],
                                                                                          self.__complexes[i])['label'],
                                                                   positive=True)

    def __create_b_matrix(self):
        # (bt = NullSpace[Transpose[Y.A]]) // MatrixForm
        the_null_space = (self.__Y * self.__A).T.nullspace()

        # getting the number of columns given in nullspace computation
        sizes = len(the_null_space)

        # simplifying the entries given in the columns
        # produced by the nullspace
        for i in range(0, sizes):
            the_null_space[i] = sympy.simplify(the_null_space[i])

        bt_temp = self.__create_non_negative_b_matrix(the_null_space, sizes)
        
        # taking out zero rows if they are in unique vectors
        bt = sympy.Matrix(bt_temp[~numpy.all(bt_temp == 0, axis=1)]).evalf(10)

        # making the dimension of bt be lambda by N, just incase
        # we have more vectors than we should
        bt = bt[0:self.get_dim_equilibrium_manifold(), :]

        # putting in a check to make sure that the rows
        # of bt are linearly independent
        _, vals = bt.T.rref()
        if len(vals) != self.get_dim_equilibrium_manifold():
            print(" ")
            print("Error: Not all rows of B^T are linearly independent!")
            print(" ")
            sys.exit()

        self.__B = bt

    def __create_non_negative_b_matrix(self, the_null_space, sizes):

        a_null = numpy.zeros((len(self.__species), sizes))
        for i in range(len(self.__species)):
            temp_vec = []
            for j in range(sizes):
                temp_vec.append(the_null_space[j][i])
            a_null[i, :] = temp_vec

        a_eq = numpy.array([numpy.sum(a_null, axis=0)])
        # must multiply by negative one because in optimization we have the inequality <= 0.0
        a_ub = -1.0*a_null
        b_ub = numpy.zeros(len(self.__species))
        b_eq = numpy.array([1.0])        
        
        # defining the number of solutions to simulate
        num_sols = ((a_ub.shape[0]+1)*(a_ub.shape[1]+1))*10

        # a matrix to hold the different solutions of the method
        sol = numpy.zeros((num_sols, len(self.__species)))

        # searching the nullspace for nonnegative vectors
        for i in range(num_sols):
            # generating a multivariate normal random variates
            crit = numpy.random.normal(0, 1.0, a_ub.shape[1])

            # Solving the linear programming problem
            out = scipy.optimize.linprog(crit, A_eq=a_eq, b_eq=b_eq, A_ub=a_ub, b_ub=b_ub, bounds=(None, None),
                                         method='simplex')
            # multiplying our nullspace vectors with the minimized value
            # to create a nonnegative vector
            sol[i, :] = numpy.dot(a_null, out.x)
            sol[i, [numpy.abs(sol[i, :]) < numpy.finfo(float).eps][0]] = numpy.float64(0.0)

        # getting the unique vectors
        unique_vecs = numpy.unique(sol, axis=0)
        num_rows = numpy.size(unique_vecs, 0)

        # taking the smallest nonzero entry of the unique vectors
        # and dividing by it this will hopfully create nice
        # looking vectors that are whole numbers
        for i in range(num_rows):
            minval = numpy.min(unique_vecs[i, numpy.nonzero(unique_vecs[i, :])])
            unique_vecs[i, :] = unique_vecs[i, :]/numpy.float64(minval)
            
        unique_vecs = numpy.unique(unique_vecs.round(decimals=10), axis=0)

        return unique_vecs

    def __create_mass_action_monomials(self):
        self.__psi = sympy.ones(len(self.__complexes), 1)
        for j in range(len(self.__complexes)):
            for i in range(len(self.__species)):
                self.__psi[j] *= sympy.Symbol(self.__species[i], positive=True)**self.__Y[i, j]
                
                # TODO: possibly make concetrations positive

    # public getters
    def get_dict_id_name(self):
        return self.__dic_species_id_name

    def get_ode_system(self):
        """
        Returns SymPy matrix representing the ODE system.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the
        provided example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> sympy.pprint(network.get_c_graph().get_ode_system())
            ⎡    -re₁⋅s₁⋅s₂ + re1r⋅s₃ + re₄⋅s₁₆ - re₆⋅s₁⋅s₆ + re6r⋅s₁₅     ⎤
            ⎢                                                              ⎥
            ⎢                 -re₁⋅s₁⋅s₂ + s₃⋅(re1r + re₂)                 ⎥
            ⎢                                                              ⎥
            ⎢                 re₁⋅s₁⋅s₂ + s₃⋅(-re1r - re₂)                 ⎥
            ⎢                                                              ⎥
            ⎢re₂⋅s₃ - re₃⋅s₆⋅s₇ + re3r⋅s₁₆ - re₆⋅s₁⋅s₆ + s₁₅⋅(re6r + 2⋅re₈)⎥
            ⎢                                                              ⎥
            ⎢                -re₃⋅s₆⋅s₇ + s₁₆⋅(re3r + re₄)                 ⎥
            ⎢                                                              ⎥
            ⎢                re₆⋅s₁⋅s₆ + s₁₅⋅(-re6r - re₈)                 ⎥
            ⎢                                                              ⎥
            ⎣                re₃⋅s₆⋅s₇ + s₁₆⋅(-re3r - re₄)                 ⎦
        """
        return self.__Y*self.__A*self.__psi

    def get_graph(self):
        """
        Returns the NetworkX DiGraph representation of the network.

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/sbml_file.xml")
        >>> network.get_c_graph().get_graph()
        """
        return self.__g

    def get_species(self):
        """
        Returns Python list of strings representing the species of the network.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> print(network.get_c_graph().get_species())
            ['s1', 's2', 's3', 's6', 's7', 's15', 's16']
        """
        return self.__species

    def get_complexes(self):
        """
        Returns Python list of strings representing the complexes of the network.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> print(network.get_c_graph().get_complexes())
            ['s1+s2', 's3', 's6+s2', 's6+s7', 's16', 's7+s1', 's1+s6', 's15', '2*s6']
        """
        return self.__complexes

    def get_reactions(self):
        """
        Returns Python list of strings representing the reactions of the network.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> print(network.get_c_graph().get_reactions())
            ['re1', 're1r', 're2', 're3', 're3r', 're4', 're6', 're6r', 're8']
        """
        return self.__reactions

    def get_a(self):
        """
        Returns SymPy matrix representing the kinetic constant matrix, :math:`A`.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> sympy.pprint(network.get_c_graph().get_a())
            ⎡-re₁     re1r      0   0         0       0   0         0       0⎤
            ⎢                                                                ⎥
            ⎢re₁   -re1r - re₂  0   0         0       0   0         0       0⎥
            ⎢                                                                ⎥
            ⎢ 0        re₂      0   0         0       0   0         0       0⎥
            ⎢                                                                ⎥
            ⎢ 0         0       0  -re₃     re3r      0   0         0       0⎥
            ⎢                                                                ⎥
            ⎢ 0         0       0  re₃   -re3r - re₄  0   0         0       0⎥
            ⎢                                                                ⎥
            ⎢ 0         0       0   0        re₄      0   0         0       0⎥
            ⎢                                                                ⎥
            ⎢ 0         0       0   0         0       0  -re₆     re6r      0⎥
            ⎢                                                                ⎥
            ⎢ 0         0       0   0         0       0  re₆   -re6r - re₈  0⎥
            ⎢                                                                ⎥
            ⎣ 0         0       0   0         0       0   0        re₈      0⎦
        """
        return self.__A

    def get_y(self):
        """
        Returns SymPy matrix representing the molecularity matrix, :math:`Y`.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> sympy.pprint(network.get_c_graph().get_y())
            ⎡1  0  0  0  0  1  1  0  0⎤
            ⎢                         ⎥
            ⎢1  0  1  0  0  0  0  0  0⎥
            ⎢                         ⎥
            ⎢0  1  0  0  0  0  0  0  0⎥
            ⎢                         ⎥
            ⎢0  0  1  1  0  0  1  0  2⎥
            ⎢                         ⎥
            ⎢0  0  0  1  0  1  0  0  0⎥
            ⎢                         ⎥
            ⎢0  0  0  0  0  0  0  1  0⎥
            ⎢                         ⎥
            ⎣0  0  0  0  1  0  0  0  0⎦
        """
        return self.__Y

    def get_s(self):
        """
        Returns SymPy matrix representing the stoichiometric matrix, :math:`S`.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> sympy.pprint(network.get_c_graph().get_s())
            ⎡-1  1   0   0   0   1   -1  1   0 ⎤
            ⎢                                  ⎥
            ⎢-1  1   1   0   0   0   0   0   0 ⎥
            ⎢                                  ⎥
            ⎢1   -1  -1  0   0   0   0   0   0 ⎥
            ⎢                                  ⎥
            ⎢0   0   1   -1  1   0   -1  1   2 ⎥
            ⎢                                  ⎥
            ⎢0   0   0   -1  1   1   0   0   0 ⎥
            ⎢                                  ⎥
            ⎢0   0   0   0   0   0   1   -1  -1⎥
            ⎢                                  ⎥
            ⎣0   0   0   1   -1  -1  0   0   0 ⎦
        """
        return self.__S

    def get_b(self):
        """
        Returns SymPy matrix representing the mass conservation matrix, :math:`B`.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> sympy.pprint(network.get_c_graph().get_b())
            ⎡ 0    0    0    0   1.0   0   1.0⎤
            ⎢                                 ⎥
            ⎢ 0   1.0  1.0   0    0    0    0 ⎥
            ⎢                                 ⎥
            ⎣1.0   0   1.0  1.0   0   2.0  1.0⎦
        """
        return self.__B

    def get_lambda(self):
        """
        Returns SymPy matrix representing the linkage class matrix, :math:`\Lambda`.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> sympy.pprint(network.get_c_graph().get_lambda())
            ⎡1  0  0⎤
            ⎢       ⎥
            ⎢1  0  0⎥
            ⎢       ⎥
            ⎢1  0  0⎥
            ⎢       ⎥
            ⎢0  1  0⎥
            ⎢       ⎥
            ⎢0  1  0⎥
            ⎢       ⎥
            ⎢0  1  0⎥
            ⎢       ⎥
            ⎢0  0  1⎥
            ⎢       ⎥
            ⎢0  0  1⎥
            ⎢       ⎥
            ⎣0  0  1⎦
        """
        return self.__Lambda

    def get_psi(self):
        """
        Returns SymPy matrix representing the mass action monomials, :math:`\psi`.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> sympy.pprint(network.get_c_graph().get_psi())
            ⎡s₁⋅s₂⎤
            ⎢     ⎥
            ⎢ s₃  ⎥
            ⎢     ⎥
            ⎢s₂⋅s₆⎥
            ⎢     ⎥
            ⎢s₆⋅s₇⎥
            ⎢     ⎥
            ⎢ s₁₆ ⎥
            ⎢     ⎥
            ⎢s₁⋅s₇⎥
            ⎢     ⎥
            ⎢s₁⋅s₆⎥
            ⎢     ⎥
            ⎢ s₁₅ ⎥
            ⎢     ⎥
            ⎢   2 ⎥
            ⎣ s₆  ⎦
        """
        return self.__psi

    def get_g_nodes(self):
        """
        Returns a list of strings that represent the order of the nodes of the NetworkX DiGraph.

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/sbml_file.xml")
        >>> network.get_c_graph().get_g_nodes()
        """
        return self.__g_nodes

    def get_g_edges(self):
        """
        Returns a list of tuples of strings that represent the order of the edges of the NetworkX DiGraph.

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/sbml_file.xml")
        >>> network.get_c_graph().get_g_edges()
        """
        return self.__g_edges

    def get_weak_reversibility_of_linkage_classes(self):
        """
        Returns list of Python boolean types for the weak reversibility of each linkage class. If the linkage class is
        weakly reversible then the entry in the list is True, False otherwise with order as defined by
        :func:`crnt4sbml.Cgraph.get_linkage_classes`.

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/sbml_file.xml")
        >>> network.get_c_graph().get_weak_reversibility_of_linkage_classes()
        """
        return [networkx.is_strongly_connected(i) for i in self.__linkage_classes]

    def get_if_cgraph_weakly_reversible(self):
        """
        Returns weak reversibility of the network. If the network is weakly reversible True is returned, False otherwise.

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/sbml_file.xml")
        >>> network.get_c_graph().get_if_cgraph_weakly_reversible()
        """
        return all(self.get_weak_reversibility_of_linkage_classes())

    def get_dim_equilibrium_manifold(self):
        """
        Returns integer value representing the dimension of the equilibrium manifold, :math:`\lambda`. This value
        is the number of mass conservation relationships.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> print(network.get_c_graph().get_dim_equilibrium_manifold())
            3
        """
        return len(self.__species) - self.__S.rank()

    def get_deficiency(self):
        """
        Returns integer value representing the deficiency of the network, :math:`\delta`.
        :download:`Fig1Ci.xml <../../sbml_files/Fig1Ci.xml>` for the provided
        example.

        Example
        --------
        >>> import crnt4sbml
        >>> import sympy
        >>> network = crnt4sbml.CRNT("path/to/Fig1Ci.xml")
        >>> print(network.get_c_graph().get_deficiency())
            2
        """
        return len(self.__complexes) - len(self.__linkage_classes) - self.__S.rank()

    def get_linkage_classes(self):
        """
        Returns list of NetworkX subgraphs representing the linkage classes.

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/sbml_file.xml")
        >>> network.get_c_graph().get_linkage_classes()
        """
        return self.__linkage_classes

    def get_network_dimensionality_classification(self):
        """
        Returns a two element list specifying the dimensionality of the network.
        Possible output:
        ["over-dimensioned",0]

        or

        ["proper",1]

        or

        ["under-dimensioned",2]

        or

        ["NOT DEFINED!",3]

        Example
        --------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/sbml_file.xml")
        >>> network.get_c_graph().get_network_dimensionality_classification()
        """
        temp = self.get_dim_equilibrium_manifold() - self.get_deficiency()
        if temp < 0:
            network_class = "over-dimensioned"
            classification = 0
        elif temp == 0:
            network_class = "proper"
            classification = 1
        elif temp > 0:
            network_class = "under-dimensioned"
            classification = 2
        else:
            network_class = "NOT DEFINED!"
            classification = 3
        return [network_class, classification]

    def get_linkage_classes_deficiencies(self):
        """
        Returns an interger list of each linkage class deficiency. Here, the first element corresponds to the first
        linkage class with order as defined by :func:`crnt4sbml.Cgraph.get_linkage_classes`.

        Example
        ---------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/sbml_file.xml")
        >>> network.get_c_graph().get_linkage_classes_deficiencies()
        """
        deficiencies = []
        for l in self.__linkage_classes:
            temp = [self.__g.edges[e]['label'] for e in l.edges]
            reactions_idx = [self.__reactions.index(i) for i in temp]
            deficiencies.append(networkx.number_of_nodes(l) - 1 - self.__S[:, reactions_idx].rank())
        return deficiencies

    def get_number_of_terminal_strong_lc_per_lc(self):
        """
        Returns an integer list stating the number of terminally strong linkage classes per linkage class. Here,
        the first element corresponds to the first linkage class with order as defined by
        :func:`crnt4sbml.Cgraph.get_linkage_classes`.

        Example
        ---------
        >>> import crnt4sbml
        >>> network = crnt4sbml.CRNT("path/to/sbml_file.xml")
        >>> network.get_c_graph().get_number_of_terminal_strong_lc_per_lc()
        """
        number_of_terminal_strong_lc_per_lc = []
        for lc_i in self.__linkage_classes:
            num_term_strong = 0
            strong_lcs = [lc_i.subgraph(c).copy() for c in networkx.strongly_connected_components(lc_i)]
            for slc_j in strong_lcs:
                slc_j_lbls = [i for i in slc_j.nodes]
                x = [lc_i.successors(v_i) for v_i in slc_j_lbls]
                connected_out_vertices = [item for sublist in x for item in sublist]
                d = set(connected_out_vertices) - set(slc_j_lbls)
                if d == set():
                    num_term_strong += 1
            number_of_terminal_strong_lc_per_lc.append(num_term_strong)
        return number_of_terminal_strong_lc_per_lc

    # public methods
    def print(self):
        """
        Prints edges and nodes of NetworkX DiGraph.

        See also
        ---------
        crnt4sbml.CRNT.print_c_graph
        """
        print("")
        print("Reaction graph of the form") 
        print("reaction -- reaction label:")
        for e in self.__g_edges:
            print(e[0] + ' -> ' + e[1] + '  --  ' + self.__g.edges[e]['label'])
        print("")

    def plot(self):
        """
        Plots NetworkX DiGraph.

        See also
        ---------
        crnt4sbml.CRNT.plot_c_graph
        """
        pos = networkx.circular_layout(self.__g, scale=1.5)
        pos = networkx.kamada_kawai_layout(self.__g, pos=pos, scale=1.5)
        node_colors = [self.__g.nodes[n]['color'] for n in self.__g_nodes]
        networkx.draw(self.__g, pos, node_color=node_colors)
        node_labels = networkx.get_node_attributes(self.__g, 'label')
        networkx.draw_networkx_labels(self.__g, pos, labels=node_labels)
        edge_labels = networkx.get_edge_attributes(self.__g, 'label')
        networkx.draw_networkx_edge_labels(self.__g, pos, edge_labels, label_pos=0.3)
        plt.show()

    def plot_save(self):
        """
        Saves the plot of the NetworkX DiGraph.

        See also
        ---------
        crnt4sbml.CRNT.plot_save_c_graph
        """
        pos = networkx.circular_layout(self.__g, scale=1.5)
        pos = networkx.kamada_kawai_layout(self.__g, pos=pos, scale=1.5)
        node_colors = [self.__g.nodes[n]['color'] for n in self.__g_nodes]
        networkx.draw(self.__g, pos, node_color=node_colors)
        node_labels = networkx.get_node_attributes(self.__g, 'label')
        networkx.draw_networkx_labels(self.__g, pos, labels=node_labels)
        edge_labels = networkx.get_edge_attributes(self.__g, 'label')
        networkx.draw_networkx_edge_labels(self.__g, pos, edge_labels, label_pos=0.3)
        plt.savefig('network_cgraph.png')



# -*- coding: utf-8 -*-

"""Top-level package for CRNT4SBML."""

__author__ = """Brandon Reyes"""
__email__ = 'reyesb123@gmail.com'
__version__ = '0.0.3'

from .crnt import CRNT
from .c_graph import Cgraph
from .low_deficiency_approach import LowDeficiencyApproach
from .mass_conservation_approach import MassConservationApproach
from .semi_diffusive_approach import SemiDiffusiveApproach
from .advanced_deficiency_approach import AdvancedDeficiencyApproach


__all__ = ['CRNT', "Cgraph", "LowDeficiencyApproach", "MassConservationApproach",
           "SemiDiffusiveApproach", "AdvancedDeficiencyApproach"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

#requirements = ['Click>=6.0', ]

setup_requirements = [ ]

test_requirements = [ ]

setup(
    author="Brandon Reyes",
    author_email='reyesb123@gmail.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
    ],
    description="CRNT4SBML is a cross-platform and easily installable Python based package. CRNT4SBML is concentrated on providing a simple workflow for the testing of core CRNT methods directed at detecting bistability in cell signaling pathways endowed with mass action kinetics.",
    entry_points={
        'console_scripts': [
            'crnt4sbml=crnt4sbml.cli:main',
        ],
    },
    install_requires=['networkx>=2.3',
                      'python-libsbml>=5.18.0',
                      'numpy>=1.16.4',
                      'sympy>=1.4',
                      'scipy>=1.3.0',
                      'matplotlib>=3.1.0',
                      'antimony>=2.11.0',
                      'libroadrunner==1.5.2.1',
                      'rrplugins>=1.2.2'],
    license="Apache Software License 2.0",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    platforms = 'ALL',
    keywords='crnt4sbml',
    name='crnt4sbml',
    packages=['crnt4sbml'],
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/PNNL-Comp-Mass-Spec/CRNT4SBML',
    version='0.0.3',
    zip_safe=False,
)
